#!/usr/bin/env python3
import sys
import json
import datetime
import subprocess
import pypoolparty

qpaths = pypoolparty.slurm.testing.dummy_paths()


def job_head():
    return str.split(job_head_to_line(), "|")


def job_head_to_line():
    return "NAME|JOBID|STATE|REASON"


def job_to_line(job, delimiter="|"):
    line = ""
    head = job_head()
    for i, key in enumerate(head):
        line += job[key]
        if (i + 1) < len(head):
            line += "|"
    return line


def state_to_table(state):
    lines = []
    lines.append(job_head_to_line())
    for job in state["running"]:
        lines.append(job_to_line(job=job))
    for job in state["pending"]:
        lines.append(job_to_line(job=job))
    return str.join("\n", lines)


# dummy squeue
# ============
# Every time this is called, it runs one job.
MAX_NUM_RUNNING = 10

assert len(sys.argv) == 3
assert sys.argv[1] == "--format"
assert sys.argv[2] == "%all"

with open(qpaths["queue_state"], "rt") as f:
    state = json.loads(f.read())

evil_ichunks_num_fails = {}
evil_ichunks_max_num_fails = {}
for evil in state["evil_jobs"]:
    evil_ichunks_num_fails[evil["ichunk"]] = evil["num_fails"]
    evil_ichunks_max_num_fails[evil["ichunk"]] = evil["max_num_fails"]


if len(state["running"]) >= MAX_NUM_RUNNING:
    run_job = state["running"].pop(0)
    pypoolparty.testing.dummy_run_job(run_job)
elif len(state["pending"]) > 0:
    job = state["pending"].pop(0)
    ichunk = pypoolparty.pooling.make_ichunk_from_jobname(jobname=job["NAME"])
    if ichunk in evil_ichunks_num_fails:
        if evil_ichunks_num_fails[ichunk] < evil_ichunks_max_num_fails[ichunk]:
            job["STATE"] = "ERROR"
            state["pending"].append(job)
            evil_ichunks_num_fails[ichunk] += 1
        else:
            job["STATE"] = "RUNNING"
            state["running"].append(job)
    else:
        job["STATE"] = "RUNNING"
        state["running"].append(job)
elif len(state["running"]) > 0:
    run_job = state["running"].pop(0)
    pypoolparty.testing.dummy_run_job(run_job)


evil_jobs = []
for ichunk in evil_ichunks_num_fails:
    evil_jobs.append(
        {
            "ichunk": ichunk,
            "num_fails": evil_ichunks_num_fails[ichunk],
            "max_num_fails": evil_ichunks_max_num_fails[ichunk],
        }
    )
state["evil_jobs"] = evil_jobs


with open(qpaths["queue_state"], "wt") as f:
    f.write(json.dumps(state, indent=4))

out_table = state_to_table(state)
print(out_table)

sys.exit(0)



def get_jobs_running_pending_error(
    JB_names_set, error_state_indicator, all_jobs_running, all_jobs_pending
):
    jobs_running = filter_jobs_by_JB_name(all_jobs_running, JB_names_set)
    jobs_pending = filter_jobs_by_JB_name(all_jobs_pending, JB_names_set)
    return extract_error_from_running_pending(
        jobs_running=jobs_running,
        jobs_pending=jobs_pending,
        error_state_indicator=error_state_indicator,
    )


def extract_error_from_running_pending(
    jobs_running, jobs_pending, error_state_indicator
):
    # split into runnning, pending, and error
    _running = []
    _pending = []
    _error = []

    for job in jobs_running:
        if error_state_indicator in job["state"]:
            _error.append(job)
        else:
            _running.append(job)

    for job in jobs_pending:
        if error_state_indicator in job["state"]:
            _error.append(job)
        else:
            _pending.append(job)

    return _running, _pending, _error


def filter_jobs_by_JB_name(jobs, JB_names_set):
    my_jobs = []
    for job in jobs:
        if job["JB_name"] in JB_names_set:
            my_jobs.append(job)
    return my_jobs



import io


def make(func_module, func_name, environ, shebang=None):
    """
    Returns a string that is a python-script.
    This python-script will be executed on the worker-node.
    In here, the environment variables are set explicitly.
    It reads the chunk of tasks, runs result = func(task), and writes the
    results. The script is called on the worker-node with a single argument:

    python worker_node_script.py /path/to/work_dir/{ichunk:09d}.pkl

    Parameters
    ----------
    func_module : str
        The name of the python module containing the function to be executed.
    func_name : str
        The name of the function to be executed.
    environ : dict
        The envirionment variables to be set when the script is executed on the
        worker node.
    shebang : str (optional)
        The first line string pointing to the executable for this script.
        Example: '#!/path/to/executable'
    """
    scr = io.StringIO()
    if shebang:
        scr.write(shebang + "\n")
    scr.write("# I was generated automatically by pypoolparty.\n")
    scr.write("# I will be executed on the worker-nodes.\n")
    scr.write("import os\n")
    scr.write("import sys\n")
    scr.write("import pickle\n")
    scr.write("import pypoolparty as ppp\n")
    scr.write("import {:s}\n".format(func_module))
    scr.write("\n")
    scr.write(make_os_environ_string(environ=environ))
    scr.write("\n")
    scr.write("assert(len(sys.argv) == 2)\n")
    scr.write("chunk = ppp.utils.read_pickle(path=sys.argv[1])\n")
    scr.write("task_results = []\n")
    scr.write("for j, task in enumerate(chunk):\n")
    scr.write("    try:\n")
    scr.write(
        "        task_result = {func_module:s}.{func_name:s}(task)\n".format(
            func_module=func_module,
            func_name=func_name,
        )
    )
    scr.write("    except Exception as bad:\n")
    scr.write('        print("[task ", j, ", in chunk]", file=sys.stderr)\n')
    scr.write("        print(bad, file=sys.stderr)\n")
    scr.write("        task_result = None\n")
    scr.write("    task_results.append(task_result)\n")
    scr.write("\n")
    scr.write("ppp.utils.write_pickle(\n")
    scr.write('    path=sys.argv[1]+".out",\n')
    scr.write("    content=task_results,\n")
    scr.write(")\n")
    scr.seek(0)
    return scr.read()


def make_os_environ_string(environ):
    """
    Returns a python-code string which sets the environment variables in
    'environ'.

    Parameters
    ----------
    environ : dict
        The envirionment variables to be set when the script is executed on the
        worker node.
    """
    env = io.StringIO()
    for key in environ:
        value = environ[key]
        keydec = key.encode("unicode_escape").decode()

        if '"' in keydec:
            start = "os.environ['{key:s}']".format(key=keydec)
        else:
            start = 'os.environ["{key:s}"]'.format(key=keydec)

        iiivalue = [int(b) for b in bytes(value, "utf8")]
        stop = "str(bytes(" + str(iiivalue) + "))" + "\n"
        line = start + " = " + stop
        env.write(line)
    env.seek(0)
    return env.read()



from .version import __version__
from . import slurm
from . import sun_grid_engine
from . import chunking
from . import utils
from . import making_script
from . import pooling
from . import job_counter
from . import debugging
from . import testing



import os
from .. import utils


def dummy_paths():
    join = os.path.join
    pypoolparty_dir = utils.resources_path()
    ddir = join(pypoolparty_dir, "slurm", "tests", "resources")
    out = {}
    out["queue_state"] = join(ddir, "dummy_queue_state.json")
    out["sbatch"] = join(ddir, "dummy_sbatch.py")
    out["squeue"] = join(ddir, "dummy_squeue.py")
    out["scancel"] = join(ddir, "dummy_scancel.py")
    return out
